/* 
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/sched.h>
#include <linux/diagchar.h>
#include <mach/usbdiag.h>
#include "diagchar_hdlc.h"
#include "diagfwd.h"
#include "diagmem.h"
#include "diagchar.h"
#include "linux/oem_msg.h"

/*---------------------------------------------------------------------------
  This is the enumeration type for all  different timestamps supported.
  At this time timestamps are 8 bytes but that may change in the future
---------------------------------------------------------------------------*/
typedef enum
{
  MSG_TS_TYPE_CDMA_FULL = 0,
  MSG_TS_TYPE_WIN32,
  MSG_TS_TYPE_GW          /* WCDMA and GSM mode */
}
msg_ts_type;

typedef struct
{
  msg_hdr_type hdr;
  msg_desc_type desc;
  unsigned long args[3];
}
msg_ext_type;

/* Request for extended msg report */
#define DIAG_EXT_MSG_F            121 

#define MSG_FILE_LEN (12)           /* Allows full DOS file name + NULL */
#define MSG_FMT_LEN  (40)           /* Max len of format string + NULL */

#define MSG_GET_LOW(a,b) ((a > b) ? b : a);

#define TRUE  (1)
#define FALSE (0)
#define ON    (1)
#define OFF   (0)

#define DIAG_OEM_CMD_F          (250)
#define DIAG_OEMTEST_F          (251)
#define DIAG_SLATE_F            (246)

#define DIAG_OEM_SUB_STREAM(a) (0x0004| (a << 8))

/* TODO: These statistics needs to be shared across processes */
/* Statistics */
static unsigned long msg_drop_delta;   /* number of dropped messages */

static int msg_kernel_write(const void *buf, int size);

/* Internal function declarations */
static msg_ext_type *msg_send_prep (const msg_const_type *const_blk, unsigned int num_args, unsigned int *pLength);
static int msg_get_ssid_rt_mask(int ssid, unsigned long* mask);
static const char *msg_format_filename2 (const char *filename);

static unsigned long *p_msg_kernel_mask = NULL;

/*---------------------------------------------------------------------------------------------------
                                    Externalised functions
(Do not call any of these functions directly, use the Macros defined in msg.h instead.)
---------------------------------------------------------------------------------------------------*/
/*===========================================================================
FUNCTION MSG_SEND_3

DESCRIPTION
This function sends out a debug message with 2 arguments across DiagPkt CS interface.
Do not call directly; use macro MSG_* defined in msg.h

DEPENDENCIES
DiagPkt handle should be initialised.
===========================================================================*/
void
msg_send_3 (const msg_const_type * pconst_blk, unsigned long xx_arg1,
        unsigned long xx_arg2, unsigned long xx_arg3)
{
  msg_ext_type *pMsg = NULL;
  unsigned int nLength = 0;
  const unsigned int num_args = 3;  /* # of message arguments */

  pMsg = msg_send_prep(pconst_blk, num_args, &nLength);
  if (pMsg)
  {
    pMsg->args[0] = xx_arg1;
    pMsg->args[1] = xx_arg2;
    pMsg->args[2] = xx_arg3;
    
    msg_kernel_write((const void*) pMsg, nLength);
    diagmem_free(driver, pMsg, POOL_TYPE_COPY);
  }
} /* msg_send_3() */

static int msg_get_ssid_rt_mask(int ssid, unsigned long* mask) 
{
  int success = FALSE;
  unsigned long first;
  unsigned long last;
  unsigned long *ptr = (unsigned long *)driver->msg_masks;

  *mask = 0;
  if(p_msg_kernel_mask != NULL)
  {
    *mask = *p_msg_kernel_mask;
    return TRUE;
  }
  while (*(ptr + 1)) 
  {
    first = *ptr++;
    last  = *ptr++;
    if (ssid >= first && ssid <= last) 
    {
      p_msg_kernel_mask = (ptr + (ssid - first));
      *mask = *p_msg_kernel_mask;
      success = TRUE;
      break;
    }
    ptr += ((last - first) + 1);
  }
  return (success);
}

/*===========================================================================
FUNCTION msg_format_filename2

DESCRIPTION
retrieves the position of filename from full file path.

DEPENDENCIES
None. 
===========================================================================*/
static const char *
msg_format_filename2 (const char *filename)
{
  const char *p_front = filename;
  const char *p_end = filename + strlen (filename);

  while (p_end != p_front)
  {
    if ((*p_end == '\\') || (*p_end == ':') || (*p_end == '/'))
    {
      p_end++;
      break;
    }
    p_end--;
  }
  return p_end;
} /*const char *msg_format_filename2() */

/*===========================================================================

FUNCTION MSG_SEND_PREP

DESCRIPTION
   Prepares the buffer needed by msg_send*().
   Allocates, fills in all data except arguments, and returns a pointer
   to the allocated message buffer.  It also handles message statisitics.

RETURN VALUE
   Returns the allocated buffer, and the length of the buffer

DEPENDENCIES
   None
===========================================================================*/

static msg_ext_type *msg_send_prep(const msg_const_type *const_blk, unsigned int num_args, unsigned int *pLength)
{
  unsigned long rt_mask; 
  int valid_ssid = FALSE;
  msg_ext_type *pMsg = NULL;
  unsigned int alloc_len = 0;

  /* Check the runtime mask */
  valid_ssid = msg_get_ssid_rt_mask( (int)const_blk->desc.ss_id, &rt_mask);

  if (valid_ssid && (const_blk->desc.ss_mask & rt_mask)) 
  {
    const char* abb_filename = NULL;
    char *ptr = NULL;
    int file_len;
    int fmt_len;
    unsigned long fmt_pos = sizeof(msg_ext_type);

    abb_filename = msg_format_filename2(const_blk->fname);
    file_len = MSG_GET_LOW(strlen(abb_filename), MSG_FILE_LEN);
    fmt_len   = MSG_GET_LOW(strlen(const_blk->fmt), MSG_FMT_LEN);

    /* total number of bytes to be allocated, including dereferenced FileName and Format strings */
    alloc_len = (fmt_pos + (file_len + 1) + (fmt_len + 1));

    ptr = diagmem_alloc(driver, alloc_len, POOL_TYPE_COPY);
    pMsg = (msg_ext_type *) ptr;

    if (pMsg)
    {
      pMsg->hdr.ts_lo = 0;
      pMsg->hdr.ts_hi = 0;
      pMsg->hdr.cmd_code = DIAG_EXT_MSG_F;
      pMsg->hdr.ts_type = MSG_TS_TYPE_CDMA_FULL;
      pMsg->hdr.num_args = (unsigned char)num_args;
      pMsg->hdr.drop_cnt = (unsigned char) ((msg_drop_delta > 255) ? 255 : msg_drop_delta);
      msg_drop_delta = 0;   /* Reset delta drop count */
      pMsg->desc = const_blk->desc;
      ptr += fmt_pos;
      memcpy(ptr, (void *)(const_blk->fmt), fmt_len);
      ptr += fmt_len;
      *ptr = 0;
      ptr++;
      memcpy(ptr, (void *)(abb_filename), file_len);
      ptr += file_len;
      *ptr = 0;
    } /* if (pMsg) */
    else
    {
       alloc_len = 0;
       msg_drop_delta++;
    } /* if (pMsg) */
  }/* if (valid_ssid && (const_blk->desc.ss_mask & rt_mask)) */

  if(pLength)
  {
    *pLength = alloc_len; /* return the number of bytes allocated. */
  }
  return pMsg;
} /* msg_send_prep() */

static void msg_clear_mask(void)
{
  if(p_msg_kernel_mask != NULL)
  {
    *p_msg_kernel_mask = 0L;
  }
}

static int msg_kernel_usb_write(const void *buf, int size)
{
  int err;
  struct diag_request *usb_write_ptr = NULL;

  if (!driver->usb_connected) 
  {
    /*Drop the diag payload */
    msg_clear_mask();
    return -EIO;
  }

  usb_write_ptr = (struct diag_request *)(diagmem_alloc(driver, sizeof(struct diag_request), POOL_TYPE_WRITE_STRUCT));
  if(usb_write_ptr == NULL)
  {
    return -EIO;
  }
  usb_write_ptr->buf = (void *)buf;
  usb_write_ptr->length = size;
  err = usb_diag_write(driver->legacy_ch, usb_write_ptr);
  if (err) 
  {
    /*Free the buffer right away if write failed */
    diagmem_free(driver, usb_write_ptr, POOL_TYPE_WRITE_STRUCT);
    return -EIO;
  }
  return 0;
}
int oemmsg_sd_write(const void *buf, int size)
{
  int i;

  for (i = 0; i < driver->poolsize_write_struct; i++)
  {
    if (driver->buf_tbl[i].length == 0) 
    {
      driver->buf_tbl[i].buf = (void *)buf;
      driver->buf_tbl[i].length = size;
      break;
    }
  }
  for (i = 0; i < driver->num_clients; i++)
  {
    if (driver->client_map[i].pid == driver->logging_process_id)
    {
      break;
    }
  }
  if (i < driver->num_clients) 
  {
    driver->data_ready[i] |= MEMORY_DEVICE_LOG_TYPE;
    wake_up_interruptible(&driver->wait_q);
  } 
  else
  {
    return -EINVAL;
  }
  return 0;
}
static int msg_kernel_write(const void *buf, int size)
{
  int err = 1;
  int used;
  struct diag_send_desc_type send = { NULL, NULL, DIAG_STATE_START, 0 };
  struct diag_hdlc_dest_type enc = { NULL, NULL, 0 };
  void *buf_hdlc;

  /*First 4 bytes indicate the type of payload - ignore these */
  if( size > driver->itemsize )
  {
    size = driver->itemsize;
  }

  send.state = DIAG_STATE_START;
  send.pkt = buf;
  send.last = (void *)(buf + size - 1);
  send.terminate = 1;

  /*Allocate a buffer for CRC + HDLC framed output to USB */
  buf_hdlc = diagmem_alloc(driver, size + 8, POOL_TYPE_HDLC);
  if (!buf_hdlc) {
    driver->dropped_count++;
    return -ENOMEM;
  }

  enc.dest = buf_hdlc;
  enc.dest_last = (void *)(buf_hdlc + size + 7);

  diag_hdlc_encode(&send, &enc);

  used = (uint32_t) enc.dest - (uint32_t) buf_hdlc;

  if (driver->logging_mode == USB_MODE) 
  {
    err = msg_kernel_usb_write(buf_hdlc, used);
  }
  else if (driver->logging_mode == MEMORY_DEVICE_MODE) 
  {
    err = oemmsg_sd_write(buf_hdlc, used);
  }

  if (err) 
  {
    /*Free the buffer right away if write failed */
    diagmem_free(driver, buf_hdlc, POOL_TYPE_HDLC);
    return -EIO;
  }
  return 0;
}

int oemmsg_process_pkt(int cmd, int sub_id, int sub_cmd)
{
  if(cmd == DIAG_OEM_CMD_F)
  {
    if(sub_cmd == DIAG_OEM_SUB_STREAM(OFF))
    {
      msg_clear_mask();
    }
  }
  else if(cmd == DIAG_SLATE_F)
  {
    if((sub_id == 0x31)||(sub_id == 0x25)||(sub_id == 0x37))
    {
      return 1;
    }
  }
  return 0;
}
